import time
import unittest

from freezegun import freeze_time

from ..testutils import debugdict
from ..vfs import GC_DELETED_FILES, DatabaseType, FileSystem


class VirtualFileSystemTestCase(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.fs = FileSystem(debugdict())

        self.events_called = {"new": [], "deleted": [], "db_closed": []}

        def new_file(key, db_type, metadata):
            self.events_called["new"].append((key, db_type, metadata))

        def delete_file(key, db_type, metadata):
            self.events_called["deleted"].append((key, db_type, metadata))

        def db_closed():
            self.events_called["db_closed"].append(None)

        self.fs.add_event("new", new_file)
        self.fs.add_event("deleted", delete_file)
        self.fs.add_event("db_closed", db_closed)

    def test_add_files(self):
        with self.fs.session(current_time=500):
            self.fs.add_file(
                "important.txt", 20, 10, metadata={"metadata:some": "useful"}
            )
            self.fs.add_dir("important stuff", 20)
            self.fs.add_file("important stuff/somefile1.txt", 20, 30)

        listing = self.fs.list_dir("", depth=1).serialize()
        expected_listing = {
            "id": "",
            "attributes": {"date": 0, "path": "", "modified": 30},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "important stuff",
                    "attributes": {
                        "modified": 30,
                        "date": 20,
                        "path": "important stuff",
                    },
                    "readable": False,
                    "streamable": False,
                    "expandable": False,
                    "nested_items": [
                        {
                            "id": "somefile1.txt",
                            "attributes": {
                                "modified": 30,
                                "date": 30,
                                "size": 20,
                                "path": "important stuff/somefile1.txt",
                            },
                            "readable": True,
                            "streamable": False,
                            "expandable": False,
                            "nested_items": None,
                        }
                    ],
                },
                {
                    "id": "important.txt",
                    "attributes": {
                        "modified": 10,
                        "date": 10,
                        "size": 20,
                        "path": "important.txt",
                        "metadata:some": "useful",
                    },
                    "readable": True,
                    "expandable": False,
                    "streamable": False,
                    "nested_items": None,
                },
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("important snuff", 20)
            self.fs.add_dir("important stuff", 20)

        listing = self.fs.list_dir("", depth=1).serialize()
        expected_listing = {
            "id": "",
            "attributes": {"date": 0, "path": "", "modified": 500},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "important snuff",
                    "attributes": {
                        "modified": 20,
                        "date": 20,
                        "path": "important snuff",
                    },
                    "readable": False,
                    "expandable": False,
                    "streamable": False,
                    "nested_items": [],
                },
                {
                    "id": "important stuff",
                    "attributes": {
                        "modified": 500,
                        "date": 20,
                        "path": "important stuff",
                    },
                    "readable": False,
                    "expandable": False,
                    "streamable": False,
                    "nested_items": [],
                },
            ],
        }
        self.assertEqual(listing, expected_listing)

    def test_nested_modified(self):
        with self.fs:
            self.fs.add_dir("important stuff", 10)
            self.fs.add_file("important stuff/somefile1.txt", 20, 20)
            self.fs.add_dir("important stuff/more stuff", 30)

        listing = self.fs.list_dir("", depth=0).serialize()
        expected_listing = {
            "id": "",
            "attributes": {"date": 0, "path": "", "modified": 30},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "important stuff",
                    "attributes": {
                        "modified": 30,
                        "date": 10,
                        "path": "important stuff",
                    },
                    "readable": False,
                    "streamable": False,
                    "expandable": True,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs:
            self.fs.add_dir("important stuff/more stuff/even mores tuff", 40)

        listing = self.fs.list_dir("", depth=0).serialize()
        expected_listing = {
            "id": "",
            "attributes": {"date": 0, "path": "", "modified": 40},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "important stuff",
                    "attributes": {
                        "modified": 40,
                        "date": 10,
                        "path": "important stuff",
                    },
                    "readable": False,
                    "streamable": False,
                    "expandable": True,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs:
            self.fs.add_file(
                "important stuff/more stuff/even mores tuff/somefile1.txt", 20, 50
            )

        listing = self.fs.list_dir("", depth=0).serialize()
        expected_listing = {
            "id": "",
            "attributes": {"date": 0, "path": "", "modified": 50},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "important stuff",
                    "attributes": {
                        "modified": 50,
                        "date": 10,
                        "path": "important stuff",
                    },
                    "readable": False,
                    "streamable": False,
                    "expandable": True,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)
        self.assertEqual(
            self.fs.last_modified(
                "important stuff/more stuff/even mores tuff/somefile1.txt"
            ),
            50,
        )
        self.assertEqual(
            self.fs.last_modified("important stuff/more stuff/even mores tuff"), 50
        )
        self.assertEqual(self.fs.last_modified("important stuff/more stuff"), 50)
        self.assertEqual(self.fs.last_modified("important stuff"), 50)
        self.assertEqual(self.fs.last_modified(""), 50)

        with self.fs:
            self.fs.add_file(
                "important stuff/more stuff/even mores tuff/somefile1.txt",
                20,
                60,
                {"a": "b"},
            )

        listing = self.fs.list_dir("", depth=0).serialize()
        expected_listing = {
            "id": "",
            "attributes": {"date": 0, "path": "", "modified": 60},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "important stuff",
                    "attributes": {
                        "modified": 60,
                        "date": 10,
                        "path": "important stuff",
                    },
                    "readable": False,
                    "streamable": False,
                    "expandable": True,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)
        self.assertEqual(
            self.fs.last_modified(
                "important stuff/more stuff/even mores tuff/somefile1.txt"
            ),
            60,
        )
        self.assertEqual(
            self.fs.last_modified("important stuff/more stuff/even mores tuff"), 60
        )
        self.assertEqual(self.fs.last_modified("important stuff/more stuff"), 60)
        self.assertEqual(self.fs.last_modified("important stuff"), 60)
        self.assertEqual(self.fs.last_modified(""), 60)

    def test_update_metadata(self):
        with self.fs:
            self.fs.add_dir(
                "metadata folder", 10, {"test": "metadata", "1": 2, "list": [1, 2, 3]}
            )
            self.fs.add_file(
                "metadata folder/metadata file.txt",
                20,
                20,
                {"test2": "metadata2", "3": 4, "list2": [3, 2, 1]},
            )

        listing = self.fs.list_dir("metadata folder", depth=0).serialize()
        expected_listing = {
            "id": "metadata folder",
            "attributes": {
                "date": 10,
                "path": "metadata folder",
                "1": 2,
                "list": [1, 2, 3],
                "modified": 20,
                "test": "metadata",
            },
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "metadata file.txt",
                    "attributes": {
                        "3": 4,
                        "date": 20,
                        "path": "metadata folder/metadata file.txt",
                        "list2": [3, 2, 1],
                        "modified": 20,
                        "size": 20,
                        "test2": "metadata2",
                    },
                    "readable": True,
                    "streamable": False,
                    "expandable": False,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs:
            self.fs.add_dir(
                "metadata folder",
                10,
                {"test": "metadata2", "1": 3, "list": [1, 2, 4], "new": "value"},
            )
            self.fs.add_file(
                "metadata folder/metadata file.txt",
                20,
                20,
                {"test2": "metadata3", "3": 5, "list2": [3, 2, 3], "new": "value"},
            )

        listing = self.fs.list_dir("metadata folder", depth=0).serialize()
        expected_listing = [
            {
                "1": 3,
                "date": 10,
                "id": "metadata folder",
                "list": [1, 2, 4],
                "modified": 20,
                "name": "metadata folder",
                "test": "metadata2",
                "type": "folder",
                "new": "value",
            },
            {
                "3": 5,
                "date": 20,
                "id": "metadata folder/metadata file.txt",
                "list2": [3, 2, 3],
                "modified": 20,
                "name": "metadata file.txt",
                "size": 20,
                "test2": "metadata3",
                "type": "file",
                "new": "value",
            },
        ]
        expected_listing = {
            "id": "metadata folder",
            "attributes": {
                "1": 3,
                "date": 10,
                "list": [1, 2, 4],
                "modified": 20,
                "path": "metadata folder",
                "test": "metadata2",
                "new": "value",
            },
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "metadata file.txt",
                    "attributes": {
                        "3": 5,
                        "date": 20,
                        "list2": [3, 2, 3],
                        "modified": 20,
                        "size": 20,
                        "test2": "metadata3",
                        "new": "value",
                        "path": "metadata folder/metadata file.txt",
                    },
                    "readable": True,
                    "expandable": False,
                    "streamable": False,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs:
            self.fs.update_metadata(
                DatabaseType.DIRECTORY,
                "metadata folder",
                {"new": "value2", "horse": "ride"},
            )

        metadata = self.fs.get_metadata(DatabaseType.DIRECTORY, "metadata folder")
        expected_metadata = {
            "1": 3,
            "date": 10,
            "horse": "ride",
            "list": [1, 2, 4],
            "modified": 20,
            "new": "value2",
            "path": "metadata folder",
            "test": "metadata2",
        }
        self.assertEqual(metadata, expected_metadata)

    def test_delete_resurrect_removed(self):
        with self.fs.session(current_time=500):
            self.fs.add_dir("folder", 10)
            self.fs.add_file("folder/file", 20, 20)

        listing = self.fs.list_dir("folder", depth=0).serialize()
        expected_listing = {
            "id": "folder",
            "attributes": {"modified": 20, "date": 10, "path": "folder"},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "file",
                    "attributes": {
                        "modified": 20,
                        "date": 20,
                        "size": 20,
                        "path": "folder/file",
                    },
                    "readable": True,
                    "streamable": False,
                    "expandable": False,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)
            self.fs.add_file("folder/file", 20, 20)

        listing = self.fs.list_dir("folder", depth=0).serialize()
        self.assertEqual(listing, expected_listing)

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)

        listing = self.fs.list_dir("folder", depth=0).serialize()
        expected_listing = {
            "id": "folder",
            "attributes": {"modified": 500, "date": 10, "path": "folder"},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [],
        }
        self.assertEqual(listing, expected_listing)

        listing = self.fs.list_dir("folder", depth=0, show_deleted=True).serialize()
        expected_listing = {
            "id": "folder",
            "attributes": {"modified": 500, "date": 10, "path": "folder"},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "file",
                    "attributes": {
                        "modified": 20,
                        "date": 20,
                        "size": 20,
                        "path": "folder/file",
                        "deleted": 500,
                    },
                    "readable": True,
                    "streamable": False,
                    "expandable": False,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs.session(True, current_time=600):
            self.fs.add_dir("folder", 600)
            self.fs.add_file("folder/file", 20, 600)

        listing = self.fs.list_dir("folder", depth=0).serialize()
        expected_listing = {
            "id": "folder",
            "attributes": {"modified": 600, "date": 10, "path": "folder"},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "file",
                    "attributes": {
                        "modified": 600,
                        "date": 20,
                        "size": 20,
                        "path": "folder/file",
                    },
                    "readable": True,
                    "expandable": False,
                    "streamable": False,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)

        with self.fs.session(True, current_time=500 + GC_DELETED_FILES - 100):
            self.fs.add_dir("folder", 10)

        listing = self.fs.list_dir("folder", depth=0, show_deleted=True).serialize()
        expected_listing = {
            "id": "folder",
            "attributes": {"modified": 600, "date": 10, "path": "folder"},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "file",
                    "attributes": {
                        "modified": 600,
                        "date": 20,
                        "size": 20,
                        "path": "folder/file",
                        "deleted": 500,
                    },
                    "readable": True,
                    "expandable": False,
                    "streamable": False,
                    "nested_items": None,
                }
            ],
        }
        self.assertEqual(listing, expected_listing)

        with self.fs.session(True, current_time=500 + GC_DELETED_FILES + 100):
            self.fs.add_dir("folder", 10)

        listing = self.fs.list_dir("folder", depth=0, show_deleted=True).serialize()
        expected_listing = {
            "id": "folder",
            "attributes": {"modified": 600, "date": 10, "path": "folder"},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [],
        }
        self.assertEqual(listing, expected_listing)

    def test_last_modified(self):
        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)
            self.fs.add_file("folder/file", 20, 20)

        assert self.fs.last_modified("folder") == 20
        assert self.fs.last_modified("folder/file") == 20

        with self.fs.session(current_time=500):
            self.fs.add_file("folder/file", 30, 30)

        assert self.fs.last_modified("folder") == 30
        assert self.fs.last_modified("folder/file") == 30

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 30)

        assert self.fs.last_modified("folder") == 500

    def test_events(self):
        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)
            self.fs.add_file("folder/file", 20, 20)

        events = sorted(
            [m for (_, _, m) in self.events_called["new"]], key=lambda x: x["path"]
        )
        expected_events = [
            {"date": 10, "modified": 20, "path": "folder"},
            {"date": 20, "modified": 20, "path": "folder/file", "size": 20},
        ]
        self.assertEqual(events, expected_events)
        self.assertEqual(len(self.events_called["deleted"]), 0)
        self.events_called["new"] = []

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)
            self.fs.add_file("folder/file", 20, 20)

        self.assertEqual(len(self.events_called["new"]), 0)
        self.assertEqual(len(self.events_called["deleted"]), 0)

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)

        events = sorted(
            [m for (_, _, m) in self.events_called["deleted"]], key=lambda x: x["path"]
        )
        expected_events = [
            {
                "date": 20,
                "modified": 20,
                "path": "folder/file",
                "size": 20,
                "deleted": 500,
            }
        ]
        self.assertEqual(events, expected_events)
        self.assertEqual(len(self.events_called["new"]), 0)
        self.events_called["deleted"] = []

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)
            self.fs.add_file("folder/file", 20, 20)

        events = sorted(
            [m for (_, _, m) in self.events_called["new"]], key=lambda x: x["path"]
        )
        expected_events = [
            {"date": 20, "modified": 20, "path": "folder/file", "size": 20}
        ]
        self.assertEqual(events, expected_events)
        self.assertEqual(len(self.events_called["deleted"]), 0)
        self.events_called["new"] = []

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)
        self.events_called["deleted"] = []

        with self.fs.session(True, current_time=500 + GC_DELETED_FILES + 100):
            self.fs.add_dir("folder", 10)

        self.assertEqual(len(self.events_called["new"]), 0)
        self.assertEqual(len(self.events_called["deleted"]), 0)

    def test_add_routes(self):
        def add_file_route(item, path):
            item.add_route("file", True, False, False, kwargs={"item_id": item.id})

        def add_list_route(item, path):
            item.add_route("list", False, True, False, kwargs={"item_id": item.id})

        self.fs.add_event("file_route_needed", add_file_route)
        self.fs.add_event("list_route_needed", add_list_route)

        with self.fs.session(current_time=500):
            self.fs.add_file("a_file", 0, 0)
            self.fs.add_dir("a_folder", 0)

        listing = self.fs.list_dir("", depth=0).serialize(include_routes=True)
        expected_listing = {
            "id": "",
            "attributes": {"date": 0, "path": ""},
            "readable": False,
            "streamable": False,
            "expandable": False,
            "nested_items": [
                {
                    "id": "a_file",
                    "attributes": {
                        "modified": 0,
                        "date": 0,
                        "path": "a_file",
                        "size": 0,
                    },
                    "readable": True,
                    "expandable": False,
                    "streamable": False,
                    "nested_items": None,
                    "routes": [
                        {
                            "handler": "file",
                            "can_open": True,
                            "can_stream": False,
                            "can_list": False,
                            "priority": 0,
                            "kwargs": {"item_id": "a_file"},
                        }
                    ],
                },
                {
                    "id": "a_folder",
                    "attributes": {"modified": 0, "date": 0, "path": "a_folder"},
                    "readable": False,
                    "streamable": False,
                    "expandable": True,
                    "nested_items": None,
                    "routes": [
                        {
                            "handler": "list",
                            "can_open": False,
                            "can_stream": False,
                            "can_list": True,
                            "priority": 0,
                            "kwargs": {"item_id": "a_folder"},
                        }
                    ],
                },
            ],
            "routes": [],
        }

        self.assertEqual(listing, expected_listing)

    def test_no_depth_list_on_item(self):
        def add_list_route(item, path):
            item.expandable = True
            item.add_route("list", False, True, False, kwargs={"item_id": item.id})

        self.fs.add_event("on_item", add_list_route)

        with self.fs.session(current_time=500):
            self.fs.add_file("a_file", 0, 0)
            self.fs.add_dir("a_folder", 0)

        listing = self.fs.list_dir("a_folder", depth=-1).serialize(include_routes=True)

        expected_listing = {
            "id": "a_folder",
            "attributes": {"modified": 0, "date": 0, "path": "a_folder"},
            "readable": False,
            "streamable": False,
            "expandable": True,
            "nested_items": None,
            "routes": [
                {
                    "handler": "list",
                    "can_open": False,
                    "can_stream": False,
                    "can_list": True,
                    "priority": 0,
                    "kwargs": {"item_id": "a_folder"},
                }
            ],
        }

        self.assertEqual(listing, expected_listing)

    def test_event_update_metadata(self):
        def new_file(key, db_type, metadata):
            parent_path = "/".join(metadata["path"].split("/")[:-1])
            self.fs.update_metadata(
                DatabaseType.DIRECTORY, parent_path, {"this": "works"}
            )

        self.fs.add_event("new", new_file)

        with self.fs.session(True, current_time=500):
            self.fs.add_dir("folder", 10)
            self.fs.add_file("folder/file", 20, 20)

        metadata = self.fs.get_metadata(DatabaseType.DIRECTORY, "folder")
        self.assertEqual(metadata.get("this"), "works")



import pickle

from unplugged import Schema, fields

import leveldb

from ...plugins import DatabasePlugin


class LevelDBDatabaseSchema(Schema):
    path = fields.String()


class LevelDBDatabasePlugin(DatabasePlugin):
    plugin_name = "leveldb"

    config_schema = LevelDBDatabaseSchema
    loaded = False

    def __init__(self, config):
        self.db = leveldb.LevelDB(self.get_database_path(config["path"]))
        self.loaded = True

    def unload(self):
        self.close()

    def __getitem__(self, key):
        return pickle.loads(self.db.Get(key.encode("utf-8")))

    def __setitem__(self, key, value):
        self.db.Put(key.encode("utf-8"), pickle.dumps(value))

    def __delitem__(self, key):
        self.db.Delete(key.encode("utf-8"))

    def keys(self):
        return list(self.db.RangeIter(include_value=False))

    def __len__(self):
        return len(self.keys())

    def __contains__(self, key):
        try:
            self.db.Get(key.encode("utf-8"))
            return True
        except KeyError:
            return False

    def close(self):
        if self.loaded:
            self.loaded = False
            del self.db

    def sync(self):
        if self.loaded:
            self.db.Put(b"__donotuse", b"", sync=True)
            self.db.Delete(b"__donotuse")

    def __iter__(self):
        if self.loaded:
            return self.db.RangeIter(include_value=False)
        else:
            return []



from abc import abstractmethod

from unplugged import PluginBase


class ImageCachePlugin(PluginBase):
    plugin_type = "imagecache"

    @abstractmethod
    def get_image_path(self, url):
        """
        Return the path to the file stored locally.
        Should probably try to download it if it is not found locally
        """
        raise NotImplementedError

    @abstractmethod
    def get_image_url(self, request, url):
        """
        Create an url where the image can be reached
        """
        raise NotImplementedError



import logging

from django.db import models
from jsonfield import JSONField

from ...bases.metadata.models import BaseListingItemRelation

logger = logging.getLogger(__name__)


class EmbeddedInfo(models.Model):
    path = models.CharField(max_length=500)
    query_key = models.CharField(max_length=50)

    index = models.IntegerField(null=True)

    title = models.CharField(max_length=300, null=True)
    cover = models.URLField(null=True, max_length=1000)
    year = models.IntegerField(null=True)
    plot = models.TextField(null=True)
    duration = models.IntegerField(null=True)
    rating = models.DecimalField(null=True, decimal_places=2, max_digits=3)

    genres = JSONField(default=[])
    primary_language = models.CharField(max_length=50, null=True)

    file_source = models.CharField(null=True, max_length=100)
    group = models.CharField(max_length=300, null=True)

    mediainfo_resolution = models.CharField(null=True, max_length=100)
    mediainfo_codec = models.CharField(null=True, max_length=100)
    mediainfo_container = models.CharField(null=True, max_length=100)
    mediainfo_source = models.CharField(null=True, max_length=100)
    mediainfo_scene = models.BooleanField(default=False)
    mediainfo_dual_audio = models.BooleanField(default=False)
    mediainfo_audio = models.CharField(null=True, max_length=100)
    mediainfo_best = models.BooleanField(
        default=False
    )  # probably the best choice if you have to choose

    bittorrent_seeders = models.IntegerField(null=True)
    bittorrent_leechers = models.IntegerField(null=True)
    bittorrent_snatched = models.IntegerField(null=True)

    episodeinfo_episode_type = models.CharField(max_length=200, blank=True, null=True)

    episodeinfo_season = models.IntegerField(blank=True, null=True)
    episodeinfo_episode = models.IntegerField(blank=True, null=True)
    episodeinfo_year = models.IntegerField(blank=True, null=True)
    episodeinfo_month = models.IntegerField(blank=True, null=True)
    episodeinfo_day = models.IntegerField(blank=True, null=True)
    episodeinfo_sub_title = models.CharField(max_length=150, blank=True, null=True)

    datetime = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (("query_key", "path"),)

    @property
    def metadata_name(self):
        return "embedded"

    @property
    def identifier(self):
        return self.pk

    def set_available(self):
        if self.file_source == "bittorrent":
            self.bittorrent_available = True
            self.save()


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(EmbeddedInfo, on_delete=models.CASCADE)



from django.db import models
from jsonfield import JSONField


class ListingCache(models.Model):
    app = models.CharField(max_length=30)
    searcher_name = models.CharField(max_length=50)
    search_token = models.CharField(max_length=50)
    path = models.CharField(max_length=500, default="")
    listing = JSONField()
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = (("app", "searcher_name", "search_token", "path"),)



import logging
from threading import Lock

from imdbparser import IMDb  # NOQA
from rest_framework import serializers
from unplugged import fields

from ...bases.metadata.remotemetadata import (
    MetadataSerializer,
    RemoteMetadataHandlerPlugin,
)
from ...bases.metadata.schema import MetadataSchema
from .filters import MetadataFilter
from .models import ListingItemRelation, Metadata, MetadataResolutionLink

logger = logging.getLogger(__name__)


class IMDBMetadataSchema(MetadataSchema):
    search_resolve = fields.String(
        enum=["tv", "movie"], enum_names=["TV", "Movie"]
    )  # choices: tv, movies


class IMDBMetadataSerializer(MetadataSerializer):
    primary_language = serializers.StringRelatedField()
    genres = serializers.StringRelatedField(many=True)
    languages = serializers.StringRelatedField(many=True)
    countries = serializers.StringRelatedField(many=True)
    actors = serializers.StringRelatedField(many=True)
    writers = serializers.StringRelatedField(many=True)
    directors = serializers.StringRelatedField(many=True)

    class Meta(MetadataSerializer.Meta):
        model = Metadata


class IMDBMetadataHandlerPlugin(RemoteMetadataHandlerPlugin):
    plugin_name = "imdb"
    select_related = ["primary_language"]
    prefetch_related = [
        "genres",
        "alternativetitle_set",
        "languages",
        "countries",
        "actors",
        "writers",
        "directors",
    ]
    currently_updating_lock = Lock()
    serializer = IMDBMetadataSerializer
    model = Metadata
    filter = MetadataFilter
    listing_item_relation_model = ListingItemRelation
    metadata_resolution_link_model = MetadataResolutionLink
    fulltext_fields = ["title"]

    config_schema = IMDBMetadataSchema

    @property
    def __traits__(self):
        traits = ["primary_metadata"]
        if self.config.get("search_resolve"):
            traits.append(f"metadata_{self.config['search_resolve']}")
        return traits



import hashlib
import logging
from contextlib import contextmanager
from datetime import timedelta

from asgiref.sync import async_to_sync
from django.utils.timezone import now
from thomas import Item
from unplugged import command

from ...exceptions import PathNotFoundException
from ...plugins import SearcherFilter
from ...utils import urljoin
from .models import AvailableMedia, ListingCache, TorrentFile, WebSession

logger = logging.getLogger(__name__)


class FilterRewriteMixin:  # trident: site mapping
    field_rewrite = {}
    choices_rewrite = {}  # trident-field: {trident-choice: rewrite target}
    order_rewrite = {}

    postprocess_field = {}  # target field, what to do with various stuff

    @staticmethod
    def postprocess_flatten(joiner=","):
        def postprocess(v):
            if isinstance(v, list):
                return joiner.join(v)
            else:
                return v

        return postprocess

    @staticmethod
    def postprocess_pickfirst(items):
        if items:
            return items[0]

    def parse_search_query(self, search_query):
        result = {}

        for k, v in list(search_query.items()) + [("o", search_query.order_by)]:
            if k in self.choices_rewrite:
                if not isinstance(v, list):
                    v = [v]

                target_choices = []
                usable_choices = self.choices_rewrite[k]
                for choice in v:
                    c = usable_choices.get(choice)
                    if not c:
                        continue
                    target_choices.append(c)

                v = target_choices

            f = self.field_rewrite.get(k)
            if not f or not v:
                continue

            do_postprocess = True
            if isinstance(f, dict):
                do_postprocess = False
                for fk, fv in f.items():
                    if fv == "+":
                        f = fk
                        do_postprocess = True
                    else:
                        result[fk] = fv

            if f in self.postprocess_field and do_postprocess:
                v = self.postprocess_field[f](v)

            result[f] = v

        return self.postprocess_search_query(result)

    def postprocess_search_query(self, search_query):
        return search_query

    @property
    def filters(self):
        sf = SearcherFilter(self.field_rewrite.keys())

        for field, choices in self.choices_rewrite.items():
            sf.set_choices(field, choices.keys())

        for field in self.order_rewrite.keys():
            sf.add_order_by(field)

        return sf


class LoginMixin:
    base_url = None

    login_failure_delay = timedelta(hours=2)
    login_path = None
    login_test_path = None
    login_fields = (
        None  # list of fields to take from config, if tuple, use mapped field.
    )

    last_failure = None
    login_checked = False
    last_login_check = None
    login_check_delay = timedelta(hours=2)

    def check_site_usable(self, session):
        logger.info(f"Checking if {self.name} is usable")
        if (
            self.login_checked
            and self.last_failure
            and self.last_failure + self.login_failure_delay > now()
        ):
            raise PathNotFoundException()

        if self.login_checked or self.login_test(session):
            self.login_checked = True
            return

        self.login(session)

    def login(self, session):
        logger.info(f"Clearing session and logging in to {self.name}")
        session.cookies.clear()
        data = {}
        for field in self.login_fields:
            if isinstance(field, (list, tuple)):
                field, form_field = field
                if form_field.startswith("+"):
                    data[form_field[1:]] = self.config[field]
                else:
                    data[field] = form_field
            else:
                data[field] = self.config[field]

        url = urljoin(self.base_url, self.login_path)

        r = session.post(url, data=data, allow_redirects=False)

        if r.status_code not in [200, 302] or not self.login_test(session):
            logger.warning("Failed to login")
            self.last_failure = now()
            self.login_checked = False
            raise PathNotFoundException()

    def login_test(self, session):
        logger.info(f"Testing login at {self.name}")

        if (
            self.last_login_check
            and self.last_login_check + self.login_check_delay > now()
        ):
            return True

        url = urljoin(self.base_url, self.login_test_path)
        r = session.get(url, allow_redirects=False)
        if r.status_code == 200:
            self.last_login_check = now()
            return True
        else:
            self.last_login_check = None
            return False


class WebSessionMixin:
    @contextmanager
    def get_session(self):
        with WebSession.objects.get_session(self.name) as session:
            yield session


class TorrentMixin:
    last_failure = None

    def fetch_torrent_data(self, session, url):
        return session.get(url)

    def fetch_torrent(self, url):
        url_hash = hashlib.sha1(url.encode("utf-8")).hexdigest()
        try:
            tf = TorrentFile.objects.get(app=self.name, url_hash=url_hash)
        except TorrentFile.DoesNotExist:
            if url.startswith("magnet"):
                magnet_resolver = self.config.get("magnet_resolver")
                if not magnet_resolver:
                    raise PathNotFoundException()

                torrent_data = magnet_resolver.magnet_to_torrent(url)
                if not torrent_data:
                    raise PathNotFoundException()
            else:
                try:
                    with self.get_session() as session:
                        r = self.fetch_torrent_data(session, url)
                except:
                    logger.exception("Failed to fetch torrent")
                    raise PathNotFoundException()
                if r.status_code != 200:
                    self.last_failure = now()
                    raise PathNotFoundException()
                torrent_data = r.content

            tf = TorrentFile.objects.create(
                app=self.name, url_hash=url_hash, url=url, torrent_data=torrent_data
            )

        return tf

    def stream_torrent(self, client, torrent_file, path=None):
        infohash = torrent_file.get_infohash()

        return client.stream(infohash, torrent_file.torrent_data, path=path)

    def try_decode(self, value):
        try:
            return value.decode("utf-8")
        except UnicodeDecodeError:
            logger.debug(f"Failed to decode {value} using UTF-8")

        return value.decode("iso-8859-1")

    def is_legal_path(self, path):
        for p in path:
            if p in [".", ".."] or "/" in p:
                return False
        return True

    def list_torrent(self, item, torrent_file):
        torrent = torrent_file.get_torrent_data()

        availability_metadata = item.get("metadata:availability")

        def get_folder(folders, item, path):
            if not path:
                return item

            if tuple(path) in folders:
                return folders[tuple(path)]

            parent_item = get_folder(folders, item, path[:-1])
            folders[tuple(path)] = folder_item = Item(id=path[-1])
            parent_item.add_item(folder_item)
            return folder_item

        if b"files" in torrent[b"info"]:  # multifile torrent
            folders = {}
            for f in torrent[b"info"][b"files"]:
                logger.trace("Handling torrent file %r" % (f,))
                path = [
                    self.try_decode(x) for x in f[b"path"] if x
                ]  # remove empty fragments
                if not self.is_legal_path(path):
                    logger.warning(f"Dangerous path {path!r} found, skipping")
                    continue

                name = path.pop()
                torrent_item = Item(id=name, attributes={"size": f[b"length"]})
                if availability_metadata:
                    torrent_item["metadata:availability"] = availability_metadata

                folder_item = get_folder(folders, item, path)
                folder_item.add_item(torrent_item)
        else:
            name = self.try_decode(torrent[b"info"][b"name"])
            if self.is_legal_path(name):
                torrent_item = Item(
                    id=name, attributes={"size": torrent[b"info"][b"length"]}
                )
                if availability_metadata:
                    torrent_item["metadata:availability"] = availability_metadata

                item.add_item(torrent_item)

        return item


class CacheSearchMixin:
    search_result_cache_time = timedelta(minutes=15)

    @command(
        name="clear_listing_cache",
        display_name="Clear listing caches",
        description="Removes listing caches for this plugin and forces a rebuild. This can break exisiting nested listings",
    )
    def clear_listing_cache(self):
        ListingCache.objects.filter(app=self.name).delete()

    def get_cache_results(self, search_token, path):
        try:
            listing_cache = ListingCache.objects.get(
                app=self.name, search_token=search_token, path=path
            )
        except ListingCache.DoesNotExist:
            logger.info(f"Unable to find listing for search_token:{search_token}")
            return

        return Item.unserialize(listing_cache.listing)

    def cache_search_result(self, search_token, path, listing):
        ListingCache.objects.filter(
            app=self.name, search_token=search_token, path=path
        ).delete()
        ListingCache(
            app=self.name,
            search_token=search_token,
            path=path,
            listing=listing.serialize(include_routes=True),
        ).save()



import logging

from django.apps import AppConfig as DjangoAppConfig

logger = logging.getLogger(__name__)


class AppConfig(DjangoAppConfig):
    name = "tridentstream.indexers.whoosh"
    verbose_name = "Whoosh Indexer"
    label = "indexers_whoosh"

    def ready(self):
        try:
            from .handler import WhooshIndexerPlugin  # NOQA
        except ImportError:
            logger.warning("Unable to load whoosh index plugin, please install whoosh")



from abc import abstractmethod

from unplugged import PluginBase


class TagPlugin(PluginBase):
    plugin_type = "tag"

    @abstractmethod
    def tag_listingitem(self, config, user, listingitem, tag_name):
        """
        Tag a listingitem and its linkable metadata.
        """
        raise NotImplementedError

    @abstractmethod
    def untag_listingitem(self, config, user, listingitem, tag_name):
        """
        Untag a listingitem and its linkable metadata.
        """
        raise NotImplementedError



import rest_framework_filters as filters
from django.db.models import Max, Min

from .models import ListingItemRelation


class MetadataFilter(filters.FilterSet):
    # metadata = filters.DateTimeFilter(name='firstseen__metadata')
    datetime = filters.DateTimeFilter(
        field_name="metadata__datetime", name="metadata__datetime"
    )

    class Meta:
        model = ListingItemRelation
        order_by = ["datetime"]
        order_by_aggregate = {"datetime": (Max, Min)}
        fields = []



import mimetypes
import os

from django.core.signing import BadSignature, Signer
from django.http import Http404, HttpResponse
from PIL import Image as PImage
from rest_framework import permissions
from rest_framework.views import APIView

mimetypes.init()


class ImageCacheView(APIView):
    service = None

    permission_classes = (permissions.AllowAny,)

    def get(self, request, imgcache_name):
        url = request.GET.get("url")
        if not url:
            raise Http404

        signer = Signer()
        try:
            url = signer.unsign(url)
        except BadSignature:
            raise Http404

        for image_cache in self.service._related_plugins:
            if image_cache.name == imgcache_name:
                break
        else:
            raise Http404

        path = image_cache.get_image_path(url)
        if not path:
            raise Http404

        if "resize" in request.GET:
            resized_path = path.split(".")
            resized_path[-2] = resized_path[-2] + "_resized"
            resized_path = ".".join(resized_path)

            if not os.path.isfile(resized_path):
                i = PImage.open(path)

                if i.size[0] > 640 or i.size[1] > 640 * 2:
                    i.thumbnail((640, 640 * 2), PImage.ANTIALIAS)
                    i.save(resized_path, i.format)
                else:
                    os.link(path, resized_path)

            path = resized_path

        content_type, _ = mimetypes.guess_type(path)
        response = HttpResponse(content_type=content_type)
        response["X-Sendfile"] = path

        return response



import logging

from django.apps import AppConfig as DjangoAppConfig

logger = logging.getLogger(__name__)


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.name"
    verbose_name = "Name Metadata"
    label = "metadata_name"

    def ready(self):
        from .handler import NameMetadataHandlerPlugin  # NOQA



import logging
import time
from datetime import timedelta

import requests
from django.utils.timezone import now
from thomas import Item, router

from ...exceptions import PathNotFoundException, StreamFailedException
from ...plugins import SearcherPlugin
from ...utils import urljoin
from .mixins import (
    CacheSearchMixin,
    FilterRewriteMixin,
    LoginMixin,
    TorrentMixin,
    WebSessionMixin,
)
from .schema import LoginTorrentSearcherSchema, TorrentSearcherSchema, URLSearcherSchema

logger = logging.getLogger(__name__)


class URLSearcherPluginBase(
    CacheSearchMixin, FilterRewriteMixin, WebSessionMixin, SearcherPlugin
):
    base_url = None
    search_path = None
    config_schema = URLSearcherSchema

    def __init__(self, config):
        self.config = config
        self.priority = config.get("priority", 10)
        self.route_searcher_base_query = f"searcher_base_query_{self.name}"
        self.route_searcher_base_list = f"searcher_base_list_{self.name}"
        self.route_searcher_base_stream = f"searcher_base_stream_{self.name}"

        router.register_handler(
            self.route_searcher_base_query, self.thomas_query, False, True, False
        )
        router.register_handler(
            self.route_searcher_base_list, self.thomas_list, False, True, False
        )
        router.register_handler(
            self.route_searcher_base_stream, self.thomas_stream, False, False, True
        )

    def build_url(self, url, params=None):
        return (
            requests.Request("GET", urljoin(self.base_url, url), params=(params or {}))
            .prepare()
            .url
        )

    def get_item(self, search_token, search_query):
        query = self.parse_search_query(search_query)
        item = Item(id="")
        item.expandable = True
        kwargs = {
            "url": self.build_url(self.search_path),
            "query": query,
            "search_token": search_token,
            "path": "",
            "response_type": "root",
        }
        item.add_route(
            self.route_searcher_base_query,
            False,
            True,
            False,
            kwargs=kwargs,
            priority=self.priority,
        )

        return item

    def do_query(self, session, url, query, headers=None):
        return session.get(self.build_url(url, query), headers=headers)

    def add_cache_list_routes(self, item, search_token, root_path, path="", skip=False):
        if path:
            path = f"{path}/{item.id}"
        else:
            path = item.id

        if item.is_expanded:
            for nested_item in item.nested_items:
                self.add_cache_list_routes(nested_item, search_token, root_path, path)

            if not skip:
                item.expandable = True
                item.add_route(
                    self.route_searcher_base_list,
                    False,
                    True,
                    False,
                    kwargs={
                        "path": root_path,
                        "sub_path": path,
                        "search_token": search_token,
                    },
                )

    def add_stream_routes(
        self, item, stream_info, path="", skip=False, ensure_already_streamable=False
    ):
        if not skip:
            if path:
                path = f"{path}/{item.id}"
            else:
                path = item.id

        if not ensure_already_streamable or item.is_streamable:
            self.make_streamable(item, stream_info, path)

        if item.is_expanded:
            for nested_item in item.nested_items:
                self.add_stream_routes(nested_item, stream_info, path)

    def thomas_query(
        self,
        item,
        search_token,
        path,
        response_type,
        url,
        query,
        depth=0,
        modified_since=None,
    ):  # Fetches listing from website
        logger.debug(
            f"Trying to query search_token:{search_token} path:{path} response_type:{response_type} url:{url}"
        )

        cached_search_result = self.get_cache_results(search_token, path)
        if (
            cached_search_result
            and cached_search_result.modified > now() - self.search_result_cache_time
        ):
            logger.debug(
                f"We got a cached query result for search_token:{search_token}, using that"
            )
            return cached_search_result

        with self.get_session() as session:
            self.check_site_usable(session)

            response = self.do_query(session, url, query)

        item = self.parse_result(search_token, path, response, response_type)

        self.add_cache_list_routes(item, search_token, path, skip=True)

        item["modified"] = int(time.time())

        self.cache_search_result(search_token, path, item)

        return item

    def thomas_list(
        self, item, search_token, path, sub_path, depth=0, modified_since=None
    ):  # Queries a cache for a listing
        logger.debug(
            f"Trying to list search_token:{search_token} path:{path} sub_path:{sub_path}"
        )

        cached_search_result = self.get_cache_results(search_token, path)
        if not cached_search_result:
            logger.debug(
                f"Cannot list because we were unable to find cache for search_token:{search_token} path:{path}"
            )
            raise PathNotFoundException()

        item = cached_search_result.get_item_from_path(sub_path)
        return item

    def make_streamable(self, item, stream_info, sub_path=""):
        item.streamable = True
        item.add_route(
            self.route_searcher_base_stream,
            False,
            False,
            True,
            kwargs={"sub_path": sub_path, "stream_info": stream_info},
        )

    def make_expandable(self, item, search_token, url, query, path, response_type):
        item.expandable = True
        item.add_route(
            self.route_searcher_base_query,
            False,
            True,
            False,
            kwargs={
                "search_token": search_token,
                "url": url,
                "query": query,
                "path": path,
                "response_type": response_type,
            },
        )

    def mark_available(self, item):
        identifiers = [
            available["identifier"]
            for available in item.get("metadata:availability", [])
        ]
        availability_plugin = self.config.get("availability")
        if availability_plugin:
            for identifier in identifiers:
                availability_plugin.mark_available(self.name, identifier)

    def populate_available(self, item):
        availability_plugin = self.config.get("availability")
        if availability_plugin:
            availability_plugin.populate_available(item)

    def get_recently_downloaded_identifiers(self):
        recently_downloaded_identifiers = set()
        availability_plugin = self.config.get("availability")
        if availability_plugin:
            for available in availability_plugin.get_recently_made_available(
                self.name, now() - timedelta(hours=24)
            ):
                recently_downloaded_identifiers.add(available["identifier"])

        return list(recently_downloaded_identifiers)

    def thomas_stream(self, item, stream_info, sub_path):
        raise NotImplementedError()

    def parse_result(self, search_token, session, response, response_type):
        """
        Parse a response.
        """
        raise NotImplementedError


class TorrentStreamer:
    def __init__(self, plugin, item, torrent_info, torrent_path, priority):
        self.plugin = plugin
        self.item = item
        self.torrent_info = torrent_info
        self.torrent_path = torrent_path
        self.priority = priority

    def evaluate(self):
        return self.priority

    def stream(self):
        torrent_file = self.plugin.get_torrent(**self.torrent_info)

        self.plugin.mark_available(self.item)

        client = self.plugin.config["bittorrent_client"]
        return self.plugin.stream_torrent(client, torrent_file, self.torrent_path)


class TorrentSearcherPluginBase(TorrentMixin, URLSearcherPluginBase):
    config_schema = TorrentSearcherSchema

    def __init__(self, config):
        super().__init__(config)
        self.route_searcher_base_list_torrent = (
            f"searcher_base_list_torrent_{self.name}"
        )
        router.register_handler(
            self.route_searcher_base_list_torrent,
            self.thomas_list_torrent,
            False,
            True,
            False,
        )

    def thomas_list_torrent(
        self, item, search_token, path, torrent_info, depth=0, modified_since=None
    ):
        torrent_file = self.get_torrent(**torrent_info)

        item = self.list_torrent(item, torrent_file)

        self.add_stream_routes(item, torrent_info, skip=True)
        self.add_cache_list_routes(item, search_token, path, skip=True)
        self.cache_search_result(search_token, path, item)

        return item

    def thomas_stream(self, item, stream_info, sub_path):
        self.populate_available(item)
        is_available = item.get("metadata", {}).get("available", False)
        if (
            not is_available
            and len(self.get_recently_downloaded_identifiers())
            >= self.config.get("daily_download_count_cap", 0)
            > 0
        ):
            logger.info(
                f"Cannot stream name:{self.name} item:{item!r} because we have reached cap"
            )
            raise StreamFailedException()

        return TorrentStreamer(
            self, item, stream_info, sub_path, self.priority + (is_available * 100)
        )

    def make_expandable_torrent(self, item, search_token, path, torrent_info):
        item.expandable = True
        item.add_route(
            self.route_searcher_base_list_torrent,
            False,
            True,
            False,
            kwargs={
                "search_token": search_token,
                "path": path,
                "torrent_info": torrent_info,
            },
        )

    def get_torrent(self, url):
        return self.fetch_torrent(url)


class LoginTorrentSearcherPluginBase(LoginMixin, TorrentSearcherPluginBase):
    config_schema = LoginTorrentSearcherSchema



from ..shelve.tests import ShelfDatabaseTest
from .handler import MemoryDatabasePlugin


class MemoryDatabaseTest(ShelfDatabaseTest):
    def open_database(self):
        return MemoryDatabasePlugin({})

    def test_reopen_database(self):
        pass



from django.conf.urls import url
from unplugged import (
    DefaultPermission,
    RelatedPluginField,
    Schema,
    ServicePlugin,
    fields,
)

from ...plugins import InputPlugin
from .views import RemoteFilesystemView


class RemoteFilesystemInputSchema(Schema):
    input = RelatedPluginField(plugin_type=InputPlugin, required=True)


class RemoteFilesystemServiceSchema(Schema):
    inputs = fields.Nested(RemoteFilesystemInputSchema, many=True)


class RemoteFilesystemService(ServicePlugin):
    plugin_name = "rfs"
    config_schema = RemoteFilesystemServiceSchema
    default_permission = DefaultPermission.DENY

    def get_urls(self):
        return [url("^(?P<path>.*)", RemoteFilesystemView.as_view(service=self))]



# Generated by Django 2.2.8 on 2019-12-13 16:29

import django.db.models.deletion
import jsonfield.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("services_listing", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="EmbeddedInfo",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("path", models.CharField(max_length=500)),
                ("query_key", models.CharField(max_length=50)),
                ("index", models.IntegerField(null=True)),
                ("title", models.CharField(max_length=300, null=True)),
                ("cover", models.URLField(max_length=1000, null=True)),
                ("year", models.IntegerField(null=True)),
                ("plot", models.TextField(null=True)),
                ("duration", models.IntegerField(null=True)),
                (
                    "rating",
                    models.DecimalField(decimal_places=2, max_digits=3, null=True),
                ),
                ("genres", jsonfield.fields.JSONField(default=[])),
                ("primary_language", models.CharField(max_length=50, null=True)),
                ("file_source", models.CharField(max_length=100, null=True)),
                ("mediainfo_resolution", models.CharField(max_length=100, null=True)),
                ("mediainfo_codec", models.CharField(max_length=100, null=True)),
                ("mediainfo_container", models.CharField(max_length=100, null=True)),
                ("mediainfo_source", models.CharField(max_length=100, null=True)),
                ("mediainfo_scene", models.BooleanField(default=False)),
                ("mediainfo_best", models.BooleanField(default=False)),
                ("bittorrent_seeders", models.IntegerField(null=True)),
                ("bittorrent_leechers", models.IntegerField(null=True)),
                ("bittorrent_snatched", models.IntegerField(null=True)),
                (
                    "episodeinfo_episode_type",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                ("episodeinfo_season", models.IntegerField(blank=True, null=True)),
                ("episodeinfo_episode", models.IntegerField(blank=True, null=True)),
                ("episodeinfo_year", models.IntegerField(blank=True, null=True)),
                ("episodeinfo_month", models.IntegerField(blank=True, null=True)),
                ("episodeinfo_day", models.IntegerField(blank=True, null=True)),
                (
                    "episodeinfo_sub_title",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                ("datetime", models.DateTimeField(auto_now=True)),
            ],
            options={"unique_together": {("query_key", "path")},},
        ),
        migrations.CreateModel(
            name="ListingItemRelation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "listingitem",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_embedded",
                        to="services_listing.ListingItem",
                    ),
                ),
                (
                    "metadata",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="metadata_embedded.EmbeddedInfo",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_embedded",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
                "unique_together": {("listingitem", "metadata", "user")},
            },
        ),
    ]



import os

from setuptools import setup, find_packages

from tridentstream import __version__


EXCLUDE_FROM_PACKAGES = ['tridentstream.bin']

def readme():
    with open("README.rst") as f:
        return f.read()

setup(
    name='tridentstream',
    version=__version__,
    url='https://github.com/tridentstream/mediaserver',
    author='Anders Jensen',
    author_email='johndoee@tridentstream.org',
    description='Media Streaming Server',
    long_description=readme(),
    long_description_content_type="text/x-rst",
    license='MIT',
    packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES) + ['twisted.plugins'],
    scripts=[
        'tridentstream/bin/bootstrap-tridentstream',
        # 'tridentstream/bin/tridentstream',
    ],
    package_data={
        '': [
            'tridentstream/fixtures/tridentstream-initial.json',

            'tridentstream/services/webinterface/static/*',
            'tridentstream/services/webinterface/static/*/*',
            'tridentstream/services/webinterface/static/*/*/*',
        ],
    },
    include_package_data=True,
    install_requires=[
        'txasgiresource>=2.0.0,<2.99.99',
        'twisted>=19.10.0,<19.99.99',
        'service-identity>=18.1.0,<18.99.99',
        'channels>=2.0.0,<2.99.99',
        'Django>=2.0.0,<2.99.99',
        'pytz>=2019.3',
        'django-environ>=0.4.5,<0.4.99',
        'django-cors-headers>=3.2.0,<3.99.99',
        'django-filter>=2.2.0,<2.99.99',
        'django-picklefield>=2.0,<2.99',
        'djangorestframework>=3.11.0,<3.11.99',
        'djangorestframework-filters>=1.0.0.dev0',
        'djangorestframework-jsonapi>=3.0.0,<3.99.99',
        'jsonfield>=2.0.2,<2.99.99',
        'requests>=2.11.1',
        'rarfile>=3.1,<3.99',
        'json-rpc>=1.12.2,<1.12.99',
        'Whoosh>=2.7.4,<2.7.99',
        'guessit>=3.1.0,<3.1.99',
        'tvdb-api>=2.0,<2.999999',
        'apscheduler>=3.6.3,<3.6.99',
        'wrapt>=1.11.2,<1.11.99',
        'wampyre>=1.1.0,<1.99.99',
        'unplugged>=0.1.6,<0.1.99',
        'imdbparser>=1.0.20,<1.99.99',
        'malparser>=1.1.7,<1.99.99',
        'thomas>=2.2.3,<2.99.99',
        'timeoutthreadpoolexecutor>=1.0.2,<1.0.99',
        'certifi',
        'pillow',
    ],
    extras_require={
        'lmdb': [
            'lmdb>=0.89',
        ],
        'leveldb': [
            'leveldb>=0.193',
        ],
        'test': [
            'pytest',
            'pytest-django',
            'freezegun',
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Twisted',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.rfs"
    verbose_name = "RemoteFileSystem Service"
    label = "services_rfs"

    def ready(self):
        from .handler import RemoteFilesystemService  # NOQA



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.iteminfo"
    verbose_name = "ItemInfo Metadata"
    label = "metadata_iteminfo"

    def ready(self):
        from .handler import ItemInfoMetadataHandlerPlugin  # NOQA



# Generated by Django 2.2.8 on 2019-12-13 16:29

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("services_listing", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Availability",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(max_length=50)),
                ("identifier", models.CharField(max_length=250)),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
            options={"unique_together": {("app", "identifier")},},
        ),
        migrations.CreateModel(
            name="ListingItemAvailability",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(max_length=50)),
                ("identifier", models.CharField(max_length=250)),
                (
                    "listingitem",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="services_listing.ListingItem",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ListingItemRelation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "listingitem",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_available",
                        to="services_listing.ListingItem",
                    ),
                ),
                (
                    "metadata",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="metadata_available.Availability",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_available",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
                "unique_together": {("listingitem", "metadata", "user")},
            },
        ),
        migrations.AddIndex(
            model_name="listingitemavailability",
            index=models.Index(
                fields=["app", "identifier"], name="metadata_av_app_de4cd0_idx"
            ),
        ),
    ]



from abc import abstractmethod, abstractproperty

from unplugged import PluginBase

from ..vfs import DatabaseType


class MetadataParserPlugin(PluginBase):  # add context manager
    plugin_type = "metadataparser"

    @abstractproperty
    def pattern(self):
        """
        Regular expression to match against files.
        """
        raise NotImplementedError

    @abstractmethod
    def handle(self, vfs, virtual_path, actual_path):
        """
        Tries to find metadata from a file and set it somewhere.
        """
        raise NotImplementedError

    def set_metadata_parent_path(self, vfs, virtual_path, key, value):
        parent_path = "/".join(virtual_path.split("/")[:-1])
        metadata = {f"metadata:{self.name}:{key}": value}
        vfs.update_metadata(DatabaseType.DIRECTORY, parent_path, metadata)



import logging

from .plugin import ItemInfoPlugin  # NOQA

logger = logging.getLogger(__name__)


try:
    from .guessit import GuessItItemInfoPlugin  # NOQA
except ImportError:
    logger.warning("Unable to load GuessIt iteminfo, to fix it, please intall guessit")



import logging

from django.conf import settings
from django.db import models

from ...bases.metadata.models import BaseListingItemRelation, BaseMetadataLink

logger = logging.getLogger(__name__)


class Tag(BaseMetadataLink):
    """User who starred an item to follow it"""

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, db_index=True, on_delete=models.deletion.CASCADE
    )
    tag_name = models.CharField(max_length=30, db_index=True)
    created = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (("user", "tag_name", "content_type", "object_id"),)

    @property
    def metadata_name(self):
        return "tag"

    @property
    def identifier(self):
        return self.pk


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(Tag, on_delete=models.deletion.CASCADE)



import json
import logging

from django.http import Http404
from rest_framework import status
from rest_framework.response import Response
from unplugged import CascadingPermission, JSONAPIObject, JSONAPIRoot

from ...bases.listing.views import BaseListingView
from ...utils import hash_string
from .models import SearchQueryCache

logger = logging.getLogger(__name__)


class StoreListView(BaseListingView):
    permission_classes = (CascadingPermission,)

    def get_config_view(self, request):
        root = JSONAPIRoot()
        logger.debug(
            f"Plugin {self.service!r}/{self.service.name} with full sections config {self.service.config['sections']!r}"
        )
        for section in self.service.config["sections"]:
            logger.debug(
                f"Making config view {section['name']} with settings {section!r}"
            )

            links = {
                "self": request.build_absolute_uri(
                    f"/{self.service.name}/{section['name']}"
                )
            }
            obj = JSONAPIObject(
                "folder", f"{self.service.name}/{section['name']}", links=links
            )
            obj["name"] = section["name"]
            if section.get("display_name"):
                obj["display_name"] = section["display_name"]

            search_filter_obj = self.get_search_filter(obj["name"], section)
            if search_filter_obj:
                obj.add_relationship(
                    "metadata_searchfilter", search_filter_obj, local=True
                )

            root.append(obj)

        return root.serialize(request)

    def get_path(self, request, path):
        if "/" in path:
            return path

        query = dict(request.GET)  # TODO: use search filter to filter?
        skip_query_keys = ["limit", "page"]
        for skip_query_key in skip_query_keys:
            if skip_query_key in query:
                del query[skip_query_key]

        query_hash = hash_string(json.dumps(sorted(query.items())))
        search_query_cache, created = SearchQueryCache.objects.get_or_create(
            query_hash=query_hash, defaults={"query": query}
        )
        search_query_cache.save()

        return f"{path}/{query_hash}"

    def get_search_filter(self, name, level_config):
        obj = JSONAPIObject("metadata_searchfilter", name)

        filters = self.service.get_filters(level_config)

        if filters:
            obj["filter"] = filters.serialize()

        return obj

    def add_additional_parent_relationships(
        self, parent, request, config, listing_item_root
    ):
        path = listing_item_root.path.split("/")
        root_path, query_hash = path[:2]
        trailing_path = "/".join(path[2:])
        if (
            trailing_path
        ):  # Should not be added to nested listings, only actual searches
            return

        search_filter_obj = self.get_search_filter(root_path, config)
        parent.add_relationship("metadata_searchfilter", search_filter_obj)



from .bittorrentclient import BittorrentClientPlugin  # NOQA
from .config import ConfigPlugin  # NOQA
from .db import DatabaseCacheLayer, DatabasePlugin  # NOQA
from .history import HistoryPlugin  # NOQA
from .imagecache import ImageCachePlugin  # NOQA
from .indexer import IndexerPlugin, PathIndexer  # NOQA
from .input import InputPlugin, InputPluginManager  # NOQA
from .magnetresolver import MagnetResolverPlugin
from .metadatahandler import MetadataHandlerPlugin  # NOQA
from .metadataparser import MetadataParserPlugin  # NOQA
from .notifier import Notification, NotifierPlugin  # NOQA
from .searcher import (  # NOQA
    SearcherFilter,
    SearcherPlugin,
    SearcherPluginManager,
    SearchQuery,
)
from .tag import TagPlugin  # NOQA



import logging

import requests
from unplugged import Schema, fields
from unplugged.models import Log

from ...exceptions import PathNotFoundException
from ...plugins import BittorrentClientPlugin
from ...stream import Stream

logger = logging.getLogger(__name__)


class DelugeBittorrentClientSchema(Schema):
    url = fields.URL(required=True, require_tld=False, ui_schema={"ui:title": "URL"})
    username = fields.String(
        required=True, default="stream", ui_schema={"ui:title": "Username"}
    )
    password = fields.String(required=True, ui_schema={"ui:title": "Password"})
    label = fields.String(ui_schema={"ui:title": "Deluge Label"})

    class Meta:
        ui_schema = {"ui:order": ["url", "username", "password", "label"]}


class DelugeBittorrentClientPlugin(BittorrentClientPlugin):
    plugin_name = "deluge"
    simpleadmin_templates = True

    config_schema = DelugeBittorrentClientSchema

    def stream(self, info_hash, torrent_filelike_data, path=None):
        with Log.objects.start_chain(self, "TORRENT.ADD") as log:
            log.log(0, f"Queuing up torrent with infohash {info_hash}")

            url = self.config["url"]
            auth = (self.config["username"], self.config["password"])
            params = {
                "infohash": info_hash,
                "path": path or None,
                "wait_for_end_pieces": "1",
            }
            if self.config.get("label"):
                params["label"] = self.config["label"]

            r = requests.get(url, auth=auth, params=params).json()
            if r.get("status") != "success":
                r = requests.post(
                    url, auth=auth, params=params, data=torrent_filelike_data
                ).json()
                if r.get("status") != "success":
                    raise PathNotFoundException()

            return Stream(url=r["url"], playhandler="http", name=r["filename"])



import re

from unplugged import Schema

from ...plugins import MetadataParserPlugin


class IMDBMetadataParserPlugin(MetadataParserPlugin):
    plugin_name = "imdb"
    pattern = re.compile(r"(?i).+\.nfo$")
    config_schema = Schema

    def handle(self, vfs, virtual_path, actual_path):
        with open(actual_path, "rb") as fn:
            d = fn.read().decode("utf-8", "ignore")

        ids = re.findall(
            r"(?:(?:imdb\.com[^ ]+?(?:tt|\?|))|(?:IMDB ID: ))([0-9]{7})",
            d,
            re.IGNORECASE,
        )
        if ids:
            self.set_metadata_parent_path(vfs, virtual_path, "id", ids[0])

    def unload(self):
        pass



from django.conf.urls import url
from unplugged import (
    DefaultPermission,
    RelatedPluginField,
    Schema,
    ServicePlugin,
    fields,
)

from ...plugins import MetadataHandlerPlugin, SearcherPlugin
from .views import RemoteSearcherFiltersView, RemoteSearcherView


class RemoteSearcherSchema(Schema):
    searcher = RelatedPluginField(plugin_type=SearcherPlugin)


class RemoteSearcherServiceSchema(Schema):
    searchers = fields.Nested(RemoteSearcherSchema, many=True)
    availability = RelatedPluginField(
        plugin_type=MetadataHandlerPlugin, traits=["availability"]
    )


class RemoteSearcherService(ServicePlugin):
    plugin_name = "remotesearcher"
    config_schema = RemoteSearcherServiceSchema
    default_permission = DefaultPermission.DENY

    simpleadmin_templates = True

    def get_urls(self):
        return [
            url("^/?$", RemoteSearcherFiltersView.as_view(service=self)),
            url(
                "^(?P<searcher_name>[^/]+)/?$", RemoteSearcherView.as_view(service=self)
            ),
            url(
                "^(?P<searcher_name>[^/]+)/(?P<search_token>[^/]+)/?$",
                RemoteSearcherView.as_view(service=self),
            ),
            url(
                "^(?P<searcher_name>[^/]+)/(?P<search_token>[^/]+)/(?P<path>.+)$",
                RemoteSearcherView.as_view(service=self),
            ),
        ]



import logging

from django.db import models

from ...bases.metadata.models import BaseListingItemRelation, BaseMetadata

logger = logging.getLogger(__name__)


class Metadata(BaseMetadata):
    name = models.CharField(max_length=200)

    year = models.IntegerField(blank=True, null=True)
    season = models.IntegerField(blank=True, null=True)
    episode = models.IntegerField(blank=True, null=True)
    episode_type = models.CharField(max_length=200, blank=True, null=True)

    source = models.CharField(max_length=100, blank=True, null=True)
    fix_type = models.CharField(max_length=100, blank=True, null=True)

    @property
    def metadata_name(self):
        return "iteminfo"

    def populate(self, config):
        result = {}

        for item_info in sorted(config["item_infos"], key=lambda x: x["priority"]):
            item_info = item_info["item_info"]
            result.update(item_info.get_info(self.name))

        self.title = result.get("title")
        self.season = result.get("season")
        self.episode = result.get("episode")
        self.episode_type = result.get("episode_type")

        self.source = result.get("source")
        self.year = result.get("year")
        self.fix_type = result.get("fix_type")


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(
        Metadata, db_index=True, on_delete=models.deletion.CASCADE
    )



from abc import abstractmethod

from unplugged import PluginBase


class MagnetResolverPlugin(PluginBase):
    plugin_type = "magnetresolver"

    @abstractmethod
    def magnet_to_torrent(self, magnet_link):
        """
        Turn a magnet link into the binary torrent data.
        """



# Generated by Django 2.2.8 on 2019-12-13 16:29

import uuid

import django.db.models.deletion
import jsonfield.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("services_listing", "0001_initial"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="History",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(db_index=True, max_length=30)),
                ("name", models.CharField(db_index=True, max_length=200)),
                ("listingitem_app", models.CharField(max_length=30, null=True)),
                ("listingitem_path", models.CharField(max_length=500, null=True)),
                ("last_watched", models.DateTimeField()),
                ("created", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"ordering": ["-last_watched"],},
        ),
        migrations.CreateModel(
            name="HistoryMetadata",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                ("primary_metadata", models.BooleanField(default=False)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="metadata_history.History",
                    ),
                ),
            ],
            options={"unique_together": {("content_type", "object_id", "history")},},
        ),
        migrations.CreateModel(
            name="ViewState",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("identifier", models.UUIDField(default=uuid.uuid4, editable=False)),
                ("values", jsonfield.fields.JSONField(blank=True, default=dict)),
                ("last_update", models.DateTimeField(auto_now=True)),
                ("created", models.DateTimeField(auto_now_add=True)),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="metadata_history.History",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"unique_together": {("identifier", "user")},},
        ),
        migrations.CreateModel(
            name="ListingItemRelation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "listingitem",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_history",
                        to="services_listing.ListingItem",
                    ),
                ),
                (
                    "metadata",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="metadata_history.HistoryMetadata",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_history",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
                "unique_together": {("listingitem", "metadata", "user")},
            },
        ),
    ]



import os
from abc import abstractmethod
from collections import MutableMapping

from django.conf import settings
from unplugged import PluginBase


class DatabasePlugin(PluginBase, MutableMapping):
    plugin_type = "database"

    @abstractmethod
    def sync(self):
        """
        Flush the changes to non-volatile storage.
        """

    @abstractmethod
    def close(self):
        """
        Close the database
        """

    def get_database_path(self, path):
        """Turn a relative into an absolute path and make sure it exists"""
        if not os.path.isabs(path):
            path = os.path.join(settings.DATABASE_ROOT, path)

        if not os.path.isdir(path):
            os.makedirs(path)

        return path


class DatabaseCacheLayer(MutableMapping):
    """Caches the communication with the actual database plugin"""

    def __init__(self, db):
        self._db = db
        self._clear()

    def _clear(self):
        self._cache_tainted = set()
        self._cache = {}
        self._cache_delete = set()

    def __getitem__(self, key):
        if key in self._cache_delete:
            raise KeyError(key)

        if key not in self._cache:
            self._cache[key] = self._db[key]

        return self._cache[key]

    def __setitem__(self, key, value):
        self._cache_delete.discard(key)
        self._cache_tainted.add(key)
        self._cache[key] = value

    def __delitem__(self, key):
        if key in self._cache:
            del self._cache[key]

        self._cache_tainted.discard(key)
        self._cache_delete.add(key)

    def __iter__(self):
        return self._db.__iter__()

    def keys(self):
        keys = self._db.keys()
        keys += [k for k in self._cache.keys() if k not in set(keys)]
        return keys

    def __len__(self):
        return self._db.__len__()

    def __contains__(self, key):
        return key not in self._cache_delete and (
            key in self._cache or self._db.__contains__(key)
        )

    def _flush_cache(self):
        for k in self._cache_tainted:
            self._db[k] = self._cache[k]

        for k in self._cache_delete:
            del self._db[k]

        self._clear()

    def sync(self):
        self._flush_cache()
        return self._db.sync()

    def close(self):
        self._flush_cache()
        return self._db.close()



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.remotesearcher"
    verbose_name = "RemoteSearcher Service"
    label = "services_remotesearcher"

    def ready(self):
        from .handler import RemoteSearcherService  # NOQA



from unplugged import RelatedPluginField, Schema, fields

from ...plugins import (
    BittorrentClientPlugin,
    MagnetResolverPlugin,
    MetadataHandlerPlugin,
)


class URLSearcherSchema(Schema):
    availability = RelatedPluginField(
        plugin_type=MetadataHandlerPlugin, traits=["availability"]
    )
    priority = fields.Integer(default=10, ui_schema={"ui:title": "Priority"})


class TorrentSearcherSchema(URLSearcherSchema):
    bittorrent_client = RelatedPluginField(
        plugin_type=BittorrentClientPlugin,
        required=True,
        ui_schema={"ui:title": "Torrent client"},
    )
    daily_download_count_cap = fields.Integer(
        default=12, ui_schema={"ui:title": "Daily download cap"}
    )
    magnet_resolver = RelatedPluginField(
        plugin_type=MagnetResolverPlugin,
        required=False,
        ui_schema={"ui:title": "Resolve magnet links"},
    )


class LoginTorrentSearcherSchema(TorrentSearcherSchema):
    username = fields.String(required=True, ui_schema={"ui:title": "Username"})
    password = fields.String(
        required=True, ui_schema={"ui:widget": "password", "ui:title": "Password"}
    )



import logging
from collections import defaultdict

from rest_framework import serializers

from ...bases.metadata.linkingmetadata import LinkingMetadataHandlerPlugin
from .models import History, HistoryMetadata, ListingItemRelation, ViewState

logger = logging.getLogger(__name__)


class ViewStateSerializer(serializers.ModelSerializer):
    state = serializers.DictField(source="values")

    class Meta:
        model = ViewState
        fields = ("identifier", "state", "last_update", "created")


class HistorySerializer(serializers.ModelSerializer):
    viewstates = ViewStateSerializer(source="viewstate_set", many=True, read_only=True)

    class Meta:
        model = History
        fields = ("name", "viewstates", "last_watched", "created")


class HistoryMetadataHandlerPlugin(LinkingMetadataHandlerPlugin):
    plugin_name = "history"
    priority = -10

    serializer = HistorySerializer
    model = History
    listing_item_relation_model = ListingItemRelation
    metadata_link_model = HistoryMetadata
    user_field = "history__user_id"
    viewstate_model = ViewState
    metadata_embed_method = "relate"

    __traits__ = ["history"]

    def get_metadata(self, request, identifier):
        histories = self.model.objects.filter(user=request.user, name=identifier)
        if not histories:
            return None
        history = histories[0]

        return self.serializer(history, context={"request": request}).data

    def get_relations(self, user, listingitem_ids):
        relations = (
            self.listing_item_relation_model.objects.filter(
                listingitem__in=listingitem_ids
            )
            .filter(user=user)
            .select_related("metadata__history")
        )
        retval = defaultdict(list)
        for relation in relations:
            retval[relation.listingitem_id].append(relation.metadata.history)

        return retval

    def schedule_update(self):
        pass



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.store"
    verbose_name = "Store Service"
    label = "services_store"

    def ready(self):
        from .handler import StoreService  # NOQA



import rest_framework_filters as filters

from .models import ListingItemRelation


class MetadataFilter(filters.FilterSet):
    id = filters.AllLookupsFilter(
        field_name="metadata__identifier", name="metadata__identifier"
    )
    title = filters.AllLookupsFilter(
        field_name="metadata__title", name="metadata__title"
    )
    genre = filters.AllLookupsFilter(
        field_name="metadata__genres__name", name="metadata__genres__name"
    )  # choices
    rating = filters.AllLookupsFilter(
        field_name="metadata__rating", name="metadata__rating"
    )
    votes = filters.AllLookupsFilter(
        field_name="metadata__votes", name="metadata__votes"
    )
    duration = filters.AllLookupsFilter(
        field_name="metadata__duration", name="metadata__duration"
    )
    episodes = filters.AllLookupsFilter(
        field_name="metadata__episodes", name="metadata__episodes"
    )
    season_year = filters.AllLookupsFilter(
        field_name="metadata__season__year", name="metadata__season__year"
    )  # choices
    season_season = filters.AllLookupsFilter(
        field_name="metadata__season__season", name="metadata__season__season"
    )  # choices
    tv_rating = filters.AllLookupsFilter(
        field_name="metadata__tv_rating__name", name="metadata__tv_rating__name"
    )  # choices
    anime_type = filters.AllLookupsFilter(
        field_name="metadata__anime_type__name", name="metadata__anime_type__name"
    )  # choices

    class Meta:
        model = ListingItemRelation
        fields = []
        include_related = [
            "genre",
            "season_year",
            "season_season",
            "tv_rating",
            "anime_type",
        ]
        order_by = ["title"]



import logging

from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.auth.signals import (
    user_logged_in,
    user_logged_out,
    user_login_failed,
)
from django.db.models.signals import post_save
from unplugged.models import Log

logger = logging.getLogger(__name__)


def log_user_logged_in(sender, request, user, **kwargs):
    with Log.objects.start_chain(None, "USER.LOGIN_SUCCESS", user) as log:
        log.log(100, f"Logged in from IP {request.META.get('REMOTE_ADDR')}")


def log_user_logged_out(sender, request, user, **kwargs):
    if not user:
        return

    with Log.objects.start_chain(None, "USER.LOGOUT", user) as log:
        log.log(100, f"Logged out from IP {request.META.get('REMOTE_ADDR')}")


def log_user_login_failed(sender, credentials, request, **kwargs):
    try:
        user = User.objects.get(username=credentials.get("username", None))
    except User.DoesNotExist:
        logger.debug(
            "Unknown user %s tried to login" % (credentials.get("username", None),)
        )
        return

    with Log.objects.start_chain(None, "USER.LOGIN_FAILED", user) as log:
        log.log(100, f"Failed to login from IP {request.META.get('REMOTE_ADDR')}")


def log_user_created(sender, instance, created, **kwargs):
    if created:
        with Log.objects.start_chain(None, "USER.CREATED", instance) as log:
            log.log(100, f"Created")


def register_signals():
    user_logged_in.connect(log_user_logged_in, dispatch_uid="log_user_logged_in")
    user_logged_out.connect(log_user_logged_out, dispatch_uid="log_user_logged_out")
    user_login_failed.connect(
        log_user_login_failed, dispatch_uid="log_user_login_failed"
    )
    post_save.connect(log_user_created, sender=User, dispatch_uid="log_user_created")
